# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
pstore-lib -- Python Protected Password Store (Library)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from __future__ import absolute_import

import sys
from base64 import b64decode, b64encode
from json import load as from_jsonfile
from random import randint
from socket import error as SocketError
from sys import exc_info
from urllib import quote, unquote, urlencode
from urllib2 import HTTPError, Request, URLError, urlopen

from pstorelib.crypt import CryptoReader
from pstorelib.exceptions import (BackendDown, BackendError, NotAllowed,
                                  NotFound, NoNonce)


class Backend(object):
    """
    Interface to the PStore backend. Uses HTTP(S) as transport protocol.
    """
    def __init__(self, urls, user, verbose=False):
        self.urls = urls
        self.user = user
        self.verbose = verbose

    def newnonce(self):
        path = '/nonce.bin'
        # Setting data to the empty tuple forces a POST operation.
        try:
            reader = self._communicate(path, data=(), query={'u': self.user})
        except NotAllowed:
            raise NoNonce()
        nonce = reader.decrypt_with(None).read()
        # Passing non-UTF-8 binary data in POST data is legit, but not
        # understood by Django. We could pass it as a file, or we can encode it
        # as base64. Choosing base64 because of the limited size of the nonce.
        # Now we won't have to invent "magic" file/property names for the
        # nonce.
        return b64encode(nonce).rstrip('=')

    def propget(self, objectid, property):
        path = '/propget/%s/%s.bin' % (urlquote(objectid), urlquote(property))
        out = {'nonce_b64': self.newnonce(), 'u': self.user}
        reader = self._communicate(path, query=out)
        return reader.decrypt_with(None)

    def propset(self, objectid, property, files=None):
        assert files
        path = '/propset/%s/%s.bin' % (urlquote(objectid), urlquote(property))
        out = {'nonce_b64': self.newnonce()}
        none = self._communicate(path, data=out, files=files)
        assert none is None

    def propupd(self, objectid, files=None):
        # FIXME: add: data = new user permissions
        assert files
        path = '/propupd/%s.bin' % (urlquote(objectid),)
        out = {'nonce_b64': self.newnonce()}
        none = self._communicate(path, data=out, files=files)
        assert none is None

    def propsearch(self, allowed_only=True, propkey_icontains=None,
                   propvalue_icontains=None):
        path = '/propsearch.js'
        out = {'nonce_b64': self.newnonce()}
        if allowed_only:
            out['u'] = self.user
        if propkey_icontains is not None:
            out['propkey_icontains'] = propkey_icontains
        if propvalue_icontains is not None:
            out['propvalue_icontains'] = propvalue_icontains

        data = self._communicate(path, query=out)
        for machine, properties in data.items():
            for propkey, info in properties['properties'].items():
                if info['data']:
                    info['data'] = CryptoReader(data=b64decode(info['data']),
                                                enctype=info['enctype'])
                else:
                    del info['data']

        return data

    def validate(self):
        path = '/validate.js'
        out = {'nonce_b64': self.newnonce()}
        statusdata = self._communicate(path, query=out)
        return statusdata['errors']

    def _communicate(self, path, query=None, data=None, files=None):
        # urlencode is nice because it will accept boths dicts and lists of
        # tuples.
        query_string = ''
        if query:
            assert (isinstance(query, dict) or isinstance(query, list) or
                    isinstance(query, tuple))
            query_string = '?' + urlencode(query)

        if files:
            assert isinstance(files, list)
            form = MultiPartForm()
            if data:
                if (isinstance(data, list) or isinstance(data, tuple)):
                    iterator = data
                elif isinstance(data, dict):
                    iterator = data.iteritems()
                else:
                    assert False
                # Note that you must send ASCII or UTF-8 encoded data here.
                # Otherwise the Django POST parser will mangle the data.
                # (Therefore then nonce is sent in base64.)
                for name, value in iterator:
                    form.add_field(name, value)

            for name, filename, file in files:
                form.add_file(name, filename, file)  # no content-type..
            content_type = form.get_content_type()
            content_length = form.get_length()
            data = form.get_data()
        elif data is not None:
            assert (isinstance(data, dict) or isinstance(data, list) or
                    isinstance(data, tuple))
            content_type = 'application/x-www-form-urlencoded'
            data = urlencode(data)
            content_length = len(data)
        else:
            content_type, content_length = None, None

        return self._try_servers(path, query_string, data, content_type,
                                 content_length)

    def _try_servers(self, path, query_string, data, content_type,
                     content_length):
        first_exception = None

        for store_url in self.urls:
            if self.verbose:
                print >>sys.stderr, ('- Communicating with %s%s' %
                                     (store_url, path))

            try:
                data = self._try_server(
                    store_url, path, query_string, data, content_type,
                    content_length)

            except HTTPError, e:
                if not first_exception:
                    first_exception = exc_info()

                status_code = e.code
                try:
                    body = e.read()
                finally:
                    e.close()
                if status_code == 403:
                    raise NotAllowed()
                elif status_code == 404:
                    raise NotFound()
                elif status_code < 500:
                    if len(body) > 1024:
                        body = body[0:1021] + '...'
                    raise BackendError(body)

            except (SocketError, URLError):
                if not first_exception:
                    first_exception = exc_info()

            else:
                break
        else:
            # Re-raise original exception: connection to all stores failed.
            first_url = self.urls[0]
            exception = BackendDown('could not connect to %s' % (first_url,))
            exception.__cause__ = first_exception[1]  # PEP 3134 style
            raise exception, None, first_exception[2]  # pep8(1) complains W602

        return data

    def _try_server(self, store_url, path, query_string, data, content_type,
                    content_length):
        request = Request(str(store_url + path + query_string))
        if data is not None:
            request.add_header('Content-Type', content_type)
            request.add_header('Content-Length', content_length)
            request.add_data(data)

        file = urlopen(request)
        status = file.getcode()
        assert status in (200, 204)  # OK, OK-no-data

        # Break early if we're dealing with no-data.
        if status == 204:
            try:
                nothing = file.read()
                assert nothing == ''
            finally:
                file.close()
            data = None

        # JSON data?
        elif path.endswith('.js'):
            try:
                data = from_jsonfile(file)
            finally:
                file.close()

        # Binary data?
        elif path.endswith('.bin'):
            # We expect this to be set!
            length = int(file.headers.getheader('content-length'))
            enctype = file.headers.getheader('x-encryption', 'none')
            # No file closing here.. the cryptoreader gets to use it.
            data = CryptoReader(fp=file, length=length,
                                enctype=enctype)

        # Unknown data?
        else:
            file.close()
            raise NotImplementedError(path)

        return data

    ###########################################################################
    # UTILITY
    ###########################################################################

    def get_keys(self, users):
        path = '/users.js'
        out = [('nonce_b64', self.newnonce())] + [('q', i) for i in users]
        userdata = self._communicate(path, query=out)
        if set(users) != set(userdata.keys()):
            raise NotFound('did not receive same list of users: '
                           'got only "%s"; do the others exist?' %
                           (', '.join(userdata.keys()),))

        # Reduce dictionary to a set of keys.
        ret = {}
        for key, value in userdata.iteritems():
            ret[str(key)] = str(value['key'])
        return ret

    def get_object(self, objectid, allowed_only=True):
        path = '/object/%s.js' % (urlquote(objectid),)
        out = {'nonce_b64': self.newnonce()}
        if allowed_only:
            out['u'] = self.user
        objdata = self._communicate(path, query=out)

        # The public and shared properties are b64encoded because of the JSON
        # transport. Undo that. And add the appropriate decryption routines.
        for property, info in objdata['properties'].items():
            if info['data'] is None:
                continue

            # We're not calling decrypt_with() on it directly. The caller may
            # not want it decrypted, so we'll save us some cpu.
            info['data'] = CryptoReader(data=b64decode(info['data']),
                                        enctype=info['enctype'])

        return objdata

    def get_objects(self, objectid_contains=None, allowed_only=True,
                    verbose=False):
        nonce = self.newnonce()  # all operations require a nonce
        out = {'nonce_b64': nonce}
        if objectid_contains:
            out['q'] = objectid_contains  # e.g. a partial object name
        if allowed_only:
            out['u'] = self.user
        if verbose:
            out['v'] = '1'  # list more than just the object identifier

        path = '/objects.js'
        return self._communicate(path, query=out)


# def post_multipart(host, selector, fields, files):
#     content_type, body = encode_multipart_formdata(fields, files)
#     h = httplib.HTTP(host)
#     h.putrequest('POST', selector)
#     h.putheader('content-type', content_type)
#     h.putheader('content-length', str(len(body)))
#     h.endheaders()
#     h.send(body)
#     errcode, errmsg, headers = h.getreply()
#     return h.file.read()


class MultiPartForm(object):
    """
    Accumulate the data to be used when posting a form.

    Taken from: http://www.doughellmann.com/PyMOTW/urllib2/#uploading-files
    """
    BOUNDARY_IN = 'abcdefghijklmnopqrstuvwxyz0123456789'

    def __init__(self):
        self.form_fields = []
        self.files = []

        # Choose random boundary so that if an upload fails, it might succeed
        # on a second attempt.
        boundary = [self.BOUNDARY_IN[randint(0, len(self.BOUNDARY_IN) - 1)]
                    for i in range(48)]
        self.boundary = '--' + ''.join(boundary)

    def get_content_type(self):
        return 'multipart/form-data; boundary=%s' % self.boundary

    def add_field(self, name, value):
        """
        Add a simple field to the form data.
        """
        # Force cast to byte string so we don't incidentally upcast this to an
        # unicode string.
        self.form_fields.append((str(name), str(value)))

    def add_file(self, name, filename, file,
                 content_type='application/octet-stream'):
        """
        Add a file to be uploaded.
        """
        # Force cast to byte string so we don't incidentally upcast this to an
        # unicode string.
        # NOTE: the name is the "field name"
        # FIXME: In the future we should be able to not read the entire file
        # body at this point.
        body = file.read()
        self.files.append((str(name), str(filename), str(content_type), body))

    def get_data(self):
        """
        Return a string representing the form data, including attached files.
        """
        if not hasattr(self, '_data'):
            # Build a list of lists, each containing "lines" of the request.
            # Each part is separated by a boundary string. Once the list is
            # built, return a string where each line is separated by '\r\n'.
            parts = []
            part_boundary = '--' + self.boundary

            # Add the form fields.
            for name, value in self.form_fields:
                parts.extend([
                    part_boundary,
                    'Content-Disposition: form-data; name="%s"' % (name,),
                    '',
                    value,
                ])

            # Add the files to upload.
            for name, filename, content_type, body in self.files:
                parts.extend([
                    part_boundary,
                    ('Content-Disposition: file; name="%s"; filename="%s"' %
                     (name, filename)),
                    'Content-Type: %s' % (content_type,),
                    '',
                    body,
                ])

            # Flatten the list and add closing boundary marker, then return
            # CR+LF separated data.
            parts.extend([part_boundary + '--', ''])
            self._data = '\r\n'.join(parts)

        return self._data

    def get_length(self):
        return len(self.get_data())


def urlquote(param):
    """
    Encodes parameters for use in a path. That means that slashes get
    encoded as %2F. But then, we replace the percent sign with equals
    signs. Because of the following issue:

    http://lists.unbit.it/pipermail/uwsgi/2011-March/001621.html
    > CGI spec says that PATH_INFO should be decoded
    > WSGI says nothing, but as it does not mention REQUEST_URI as a
    > standard variable we could have no way to determine (after
    > decoding) if a / was a / or a %2F.
    """
    # Safe '' means that there are very few characters that do not get
    # escaped.
    if isinstance(param, unicode):
        param = param.encode('utf-8')
    return quote(param, safe='').replace('%', '=')


def urlunquote(param):
    """
    Undo the doings of urlquote.
    """
    param = unquote(param.replace('=', '%'))
    if isinstance(param, unicode):
        param = param.encode('latin1')  # unquote did not parse utf8
    return param.decode('utf-8')


if __name__ == '__main__':
    print 'FIXME: unittest for', __file__



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
pstore-lib -- Python Protected Password Store (Library)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from __future__ import absolute_import

import os
import re
import sys
from getpass import getpass
from sys import stderr
from time import time

try:
    from gpgme import ERR_CANCELED, Context, GpgmeError
except ImportError, e:
    raise ImportError(e.args[0] + '\n\n*HINT* apt-get install python-gpgme')

from pstorelib.bytes import BytesIO
from pstorelib.exceptions import (CryptError, CryptBadPassword,
                                  CryptBadPubKey, CryptBadPrivKey)
from pstorelib.gpgkey import get_pubkey_id_from_ascii


# A few notes about subkeys:
#
# -----------------------------------------------------------------------------
# From: http://wiki.debian.org/subkeys
#
# > GnuPG actually uses a signing-only key as the master key, and creates an
# > encryption subkey automatically. Without a subkey for encryption, you
# > can't have encrypted e-mails with GnuPG at all. Debian requires you to
# > have the encryption subkey so that certain kinds of things can be e-mailed
# > to you safely, such as the initial password for your debian.org shell
# > account.
# ...
# > You should keep your private master key very, very safe. However, keeping
# > all your keys extremely safe is inconvenient: every time you need to sign
# > a new package upload, you need to copy the packages onto suitable portable
# > media, go into your sub-basement, prove to the armed guards that you're
# > you by using several methods of biometric and other identification, go
# > through a deadly maze, feed the guard dogs the right kind of meat, and
# > then finally open the safe, get out the signing laptop, and sign they
# > packages. Then do the reverse to get back up to your Internet connection
# > for uploading the packages.
# >
# > Subkeys make this easier: you create a subkey for signing, and another for
# > encryption, and keep those on your main computer. You publish the subkeys
# > on the normal keyservers, and everyone else will use them instead of the
# > master keys, with one exception. Likewise, you will use the master keys
# > only in exceptional circumstances.
#
# -----------------------------------------------------------------------------
# From: http://serverfault.com/questions/397973/
#              gpg-why-am-i-encrypting-with-subkey-instead-of-primary-key
#
# > If you look into the details of the math of public-key encryption, you will
# > discover that signing and decrypting are actually identical operations.
# > Thus in a naive implementation it is possible to trick somebody into
# > decrypting a message by asking them to sign it.
# >
# > Several things are done in practice to guard against this. The most obvious
# > is that you never sign an actual message, instead you sign a secure hash of
# > the message. Less obviously, but just to be extra safe, you use different
# > keys for signing and encrypting.
#
# -----------------------------------------------------------------------------
# From: http://lavica.fesb.hr/cgi-bin/info2html?(gpgme)Key%20Management
#
# > The first subkey in the linked list is also called the primary key.
#
# -----------------------------------------------------------------------------
# The Key object that we get below, generally holds two subkeys:
# >>> [(i.can_sign, i.can_encrypt, i.pubkey_algo) for i in k.subkeys]
# [(True, False, 17), (False, True, 16)]
#
# (Where 17 is PUBKEY_ALGO_DSA and 16 is PUBKEY_ALGO_ELGAMAL_E.)


# Not so nice, security-wise. But without password-caching the thing becomes
# unusable now that we require the nonce for reading too.
PASSWORD_CACHE = {}


class GPGCrypt(object):
    """
    Wrapper around GPGME GPG encrypting/decrypting.
    """
    def __init__(self):
        # Gnome keyring already takes care of the key access if you're running
        # X. If you cancel the GPG popup, you get this callback instead.
        self.password_cb = self.default_password_cb
        # Init a GPG context.
        self.context = Context()
        self.context.passphrase_cb = self._password_cb  # wrap the other cb
        # No ascii armor stuff. We'll juggle some base64 around ourselves.
        self.context.armor = False

    def set_password_cb(self, password_cb):
        # Set password callback. The callback must return a (password,
        # allow_caching) tuple.
        self.password_cb = password_cb

    def generate_key(self, **kwargs):
        # <GnupgKeyParms format="internal">
        # (Sub)Key-Type: RSA
        # (Sub)Key-Length: 2048 (or 4096)
        # Name+Comment+Email "Alex Boonstra (TEST) <aboonstra@example.com>"
        # Expire-Date: None (or in a couple of years)
        # Passphrase: xxx
        # |
        # v
        # self.context.op_genkey(parms.encode('utf-8'), None, None)
        # result = self.context.op_genkey(parms.encode('utf-8'), None, None)
        msg = ('This takes too long. Generate they keys yourself. Look in '
               'docs/examples for an example')
        raise NotImplementedError(msg)

    def get_key(self, **kwargs):
        """
        Get the appropriate key to work with.

        Takes either an id= or an email= argument.

        TODO: if the key is not found, get it from server?
        """
        if (len(kwargs) != 1 or
                any(i not in ('id', 'email') for i in kwargs.keys())):
            raise TypeError('get_key takes either id= or email=')

        # Lookup by id.
        if 'id' in kwargs:
            # "If [PARAM2] is 1, only private keys will be returned."
            key = self.context.get_key(kwargs['id'], 0)
            # Make sure we have an e-mail to go by.
            email = unicode(key.uids[0].email.lower())

        # Lookup by email. E-mail returned by gpgme is in unicode. (Even
        # though it probably won't contain any non-ascii.)
        else:
            email = unicode(kwargs['email']).lower()
            found = []

            for key in self.context.keylist():
                for uid in key.uids:
                    if uid.email.lower() == email:
                        found.append(key)
                        if len(found) > 1:
                            msg = 'Cannot cope with more than one recipient'
                            raise NotImplementedError(msg)
            if not found:
                return None

            key = found[0]

        # Do checks on the key.
        if key.expired:
            raise CryptBadPubKey('%s key is expired' % (email,))
        elif key.invalid or key.revoked or key.disabled:
            raise CryptBadPubKey('%s key is invalid/revoked/diabled' %
                                 (email,))
        elif not key.can_encrypt:  # also set to false if e.g. expired
            raise CryptBadPubKey('%s key is unusable' % (email,))

        # Find right subkey and check if expiry is nigh.
        for subkey in key.subkeys:
            if not (subkey.expired or subkey.invalid or subkey.revoked or
                    subkey.disabled or not subkey.can_encrypt):
                break
        if subkey.expires and float(subkey.expires - time()) / 86400.0 < 200:
            # Send out a warning that this key is about to expires. I'm not
            # sure what the implications of expired keys are, but let's prepare
            # for the worst and warn the user at an early stage.
            print >>stderr, \
                ('WARNING: (sub)key %s for %s will expire in '
                 '%.1f days' %
                 (subkey.keyid, key.uids[0].email,
                  float(subkey.expires - time()) / 86400.0))

        # Ok. All is good.
        return key

    def import_key(self, key):
        to_import = BytesIO(str(key))  # ensure bytestring!
        res = self.context.import_(to_import)
        if (res.imported + res.unchanged) != 1:
            raise CryptError('import failed (imported=%d, unchanged=%d). '
                             'Try setting os.environ["GNUPGHOME"] to an '
                             '*existing* writable path in settings; '
                             'e.g. "/tmp"' % (res.imported, res.unchanged))

        try:
            key_id = get_pubkey_id_from_ascii(key)
            key = self.get_key(id=unicode(key_id))
        except Exception, e:
            raise CryptError('GPG quick hack failed; no key found', e)

        return key

    def decrypt(self, private_key_ref=None, input=None, output=None):
        """
        Can raise one of the CryptErrors.
        """
        assert private_key_ref is None, 'We let the keyring handle this'
        assert hasattr(input, 'read')
        assert hasattr(output, 'write')

        try:
            self.context.decrypt(input, output)
        except GpgmeError, e:
            # If you press ^C during passphrase input.
            #   gpgme.GpgmeError: (7, 58, u'No data')
            # If the decryption failed (badly encrypted, secret key missing)
            #   gpgme.GpgmeError: (7, 152, u'Decryption failed')
            # If the password callback raised an error.
            #   gpgme.GpgmeError: (7, 32779, u'Bad file descriptor')
            if e.args[0] == 7:
                if e.args[1] == 11:
                    raise CryptBadPassword()
                if e.args[1] == 152:
                    raise CryptBadPrivKey()
            raise

        # length = output.tell()
        output.seek(0)

    def encrypt(self, public_key_ref=None, input=None, output=None):
        """
        We're always encrypting to a single user at a time. That makes revoking
        user permissions easier: we don't have to re-encrypt the whole bunch.
        """
        assert public_key_ref is not None
        assert hasattr(input, 'read')
        assert hasattr(output, 'write')

        self.context.encrypt([public_key_ref], 1, input, output)

        # length = output.tell()
        output.seek(0)

    def default_password_cb(self, id_name_comment_email, key_ids,
                            prev_was_bad):
        email = re.search(r'<([^@>]+@[^>]+)>', id_name_comment_email)
        if email:
            email = email.groups()[0]
        else:
            email = '(unknown)'
        return (getpass('Enter passphrase for %s GPG key: ' % (email,)),
                True)  # allow caching

    def _password_cb(self, id_name_comment_email, key_ids, prev_was_bad, fd):
        # Password caching.
        password, allow_caching = None, False
        if id_name_comment_email in PASSWORD_CACHE:
            if prev_was_bad:
                del PASSWORD_CACHE[id_name_comment_email]
            else:
                password = PASSWORD_CACHE[id_name_comment_email]
                print >>sys.stderr, 'NOTICE: re-using cached password'

        # Get the password from the callback.
        try:
            if not password:
                password, allow_caching = self.password_cb(
                    id_name_comment_email, key_ids, bool(prev_was_bad)
                )
        except KeyboardInterrupt:
            # For some reason, we return 'No data' next..
            pass
        except Exception:
            import traceback
            traceback.print_exc()
        else:
            if password and allow_caching:
                PASSWORD_CACHE[id_name_comment_email] = password
            # Writing empty passwords too, because the close below is so
            # drastic that we wouldn't get a second try.
            assert isinstance(password, str), 'password was unicode?'
            os.write(fd, password + '\n')
            return 0

        # > The user must write the passphrase, followed by a newline
        # > character, to the file descriptor fd. If the user returns 0
        # > indicating success, the user must at least write a newline
        # > character before returning from the callback.
        # >
        # > If an error occurs, return the corresponding gpgme_error_t value.
        # > You can use the error code GPG_ERR_CANCELED to abort the operation.
        # > Otherwise, return 0.
        #
        # But that doesn't work. We must always write a newline, or the thing
        # hangs. Returning 0 or ERR_CANCELED doesn't seem to make any
        # difference. So, instead, we close() the fd. That will make for a
        # quicker abort: 'Bad file descriptor'
        os.close(fd)
        return ERR_CANCELED


if __name__ == '__main__':
    import unittest

    class Test(unittest.TestCase):
        KEY_ALEX = ('ECCF8511CA3F91BBD9BD09FEC797D1A5DD9DB7EB',
                    'alex@example.com')
        KEY_WALTER = ('ED3DB90502DE8DB34510544E66E99FD00350A01C',
                      'walter@example.com')

        def password_callback(self, id_name_comment_email, key_ids,
                              prev_was_bad):
            first_name = id_name_comment_email.rsplit('<', 1)[1]
            first_name = first_name.split('@', 1)[0]
            return (first_name + '2', False)  # (password, allow_caching)

        def test_1(self):
            g = GPGCrypt()
            g.set_password_cb(self.password_callback)

            source = 'sEcReT!'

            # FIXME/TODO: charge the GPGCrypt with the respective pgp public
            # and private keys before attempting any of this..

            recipients = (self.KEY_ALEX, self.KEY_WALTER)

            for recipient in recipients:
                key = g.get_key(id=recipient[0])

                input = BytesIO(source)
                output = BytesIO()
                g.encrypt(input=input, output=output, public_key_ref=key)
                encrypted = output.read()

                # Valid assumption when input is small.
                self.assertTrue(len(encrypted) > len(source))

                # Decrypt it
                input = BytesIO(encrypted)
                output = BytesIO()
                g.decrypt(input=input, output=output)
                decrypted = output.read()

                self.assertEquals(decrypted, source)

        # FIXME: cover more with the unit tests

    unittest.main()



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
pstore-lib -- Python Protected Password Store (Library)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from __future__ import absolute_import

from io import UnsupportedOperation
from os import fdopen
# TODO: make a SecureSpooledTemporaryFile that cleans up better after itself?
from tempfile import SpooledTemporaryFile

from pstorelib.bytes import BytesIO, can_seek, get_size, sendfile
from pstorelib.exceptions import CryptBadPrivKey


# We do late importing of pstorelib.gpg and pstorelib.sshrsa. That way you can
# get away with fetching only the libs that you need if you only use one of the
# encryption methods. (Perhaps not needed for those libs directly, but for the
# dependencies of those libs.)

# P.S. The API in this file sucks. We'll need to sort that. FIXME :)


MAX_INMEMORY_SIZE = 1048576  # 1MB or we switch to tempfiles


class PStoreCrypt(object):
    """
    XXX: document me, XXX: rename me: CryptoGlobal? CryptoStore?

    This is a singleton.
    """
    instance = None

    ASKPASS_DEFAULT = 1     # use default password callback
    ASKPASS_NEVER = 2       # do not ask for password
    ASKPASS_USERNAME2 = 3   # return username + '2' as password (for tests)

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Must be called before doing anything with it.
        """
        assert not cls.instance
        cls.instance = cls(*args, **kwargs)

    @classmethod
    def destroy(cls):
        """
        The automated test uses this. You don't need it.
        """
        del cls.instance
        cls.instance = None

    @classmethod
    def get(cls):
        """
        Get the only instance.
        """
        assert cls.instance
        return cls.instance

    def __init__(self, askpass=None, sshrsa_privkey_file=None):
        self.askpass = askpass or self.ASKPASS_DEFAULT
        self.sshrsa_privkey_file = sshrsa_privkey_file

    @property
    def sshrsa_privkey(self):
        """
        Can raise a CryptBadPassword or CryptBadPrivKey.
        """
        if not hasattr(self, '_sshrsa_privkey'):
            if isinstance(self.sshrsa_privkey_file, basestring):
                try:
                    file = open(self.sshrsa_privkey_file, 'r')
                    try:
                        key_data = file.read()
                    finally:
                        file.close()
                except Exception, e:
                    raise CryptBadPrivKey('filesystem error', e)
            else:
                try:
                    key_data = self.sshrsa_privkey_file.read()
                except Exception, e:
                    raise CryptBadPrivKey('file error', e)

            # The keyparser can ask for a password if needed.
            from pstorelib import sshrsa
            parser = sshrsa.SSHKeyParser()

            # Non-interactive use? Automated tests?
            if self.askpass == self.ASKPASS_NEVER:
                parser.set_password_cb((lambda *args: ('', False)))
            elif self.askpass == self.ASKPASS_USERNAME2:
                parser.set_password_cb((lambda name, *args: (name + '2',
                                                             False)))

            self._sshrsa_privkey = parser.parse_private(key_data)
        return self._sshrsa_privkey


class CryptoWriter(object):
    """
    XXX: document this...
    """
    def __init__(self, data=None, fp=None, length=-1):
        assert bool(data is not None) ^ bool(fp)

        if length == -1:
            if data is not None:
                length = len(data)
            else:
                length = get_size(fp)  # can be -1

        # We allow writer reuse, but if we're working with a stream, we cannot
        # seek. Copy the data to a tempfile.
        if fp and not can_seek(fp):
            newfp = SpooledTemporaryFile(MAX_INMEMORY_SIZE)
            sendfile(newfp, fp)
            length = newfp.tell()
            newfp.seek(0)
            fp = newfp

        self.data = data
        self.fp = fp
        self.fpreads = 0        # keep track of fp usage
        self.length = length

        assert length >= 0
        self.use_tempfile = (length > MAX_INMEMORY_SIZE)

    def _get_enctype(self, public_key):
        if public_key is None:
            return 'none'
        elif public_key.startswith('ssh-rsa '):
            return 'sshrsa'
        elif public_key.startswith('----'):
            return 'gpg'
        raise NotImplementedError('unknown public key type', public_key)

    def encrypt_with(self, public_key):
        """
        May raise one of the CryptErrors.
        """
        # We *could* defer encryption until the first read() call is done on
        # the outputted value. But I don't see any benefits.

        # Make sure that a reused Writer provides the right info.
        if self.fp:
            if self.fpreads > 0:
                self.fp.seek(0)  # if this breaks, we didn't init fp properly
            self.fpreads += 1

        enctype = self._get_enctype(public_key)

        if enctype == 'none':
            return self._encrypt_with_none()
        elif enctype == 'gpg':
            return self._encrypt_with_gpg(public_key)
        elif enctype == 'sshrsa':
            return self._encrypt_with_sshrsa(public_key)
        raise NotImplementedError('unknown encryption type')

    def _encrypt_with_none(self):
        if self.fp:
            # Return a new file so they're not closing ours.
            try:
                fileno = self.fp.fileno()
            except UnsupportedOperation:
                newfp = SpooledTemporaryFile(MAX_INMEMORY_SIZE)
                sendfile(newfp, self.fp)
                newfp.seek(0)
                return newfp
            else:
                return fdopen(fileno, 'rb')
        else:
            # For no encryption, passing around a BytesIO is no extra effort if
            # the data already was a string.
            return BytesIO(self.data)

    def _encrypt_with_gpg(self, public_key):
        from pstorelib import gpg
        gpgcrypt = gpg.GPGCrypt()
        gpgkey = gpgcrypt.import_key(public_key)

        if self.fp:
            input = self.fp
        else:
            input = BytesIO(self.data)
        if self.use_tempfile:
            output = SpooledTemporaryFile(MAX_INMEMORY_SIZE)
        else:
            output = BytesIO()

        gpgcrypt.encrypt(input=input, output=output, public_key_ref=gpgkey)
        return output

    def _encrypt_with_sshrsa(self, public_key):
        from pstorelib import sshrsa
        parser = sshrsa.SSHKeyParser()
        key = parser.parse_public(public_key)
        crypt = sshrsa.RSACrypt(n=key['n'], e=key['e'])

        if self.fp:
            input = self.fp.read()
        else:
            input = self.data

        # NOTE: We're not doing tempfiles here.. the sshrsa encryption method
        # is currently limited to very few bytes anyway.
        return BytesIO(crypt.encrypt(input))


class CryptoReader(object):
    """
    XXX: document this...
    """
    def __init__(self, data=None, fp=None, length=-1, enctype='none'):
        assert bool(data is not None) ^ bool(fp)

        if length == -1:
            if data is not None:
                length = len(data)
            else:
                length = get_size(fp)  # can be -1

        # We accept that we can only read it once.
        # if fp and not can_seek(fp): dont_care()

        self.data = data
        self.fp = fp
        self.fpreads = 0    # keep track of fp usage
        self.length = fp

        assert enctype in ('none', 'gpg', 'sshrsa')
        self.enctype = enctype

        self.use_tempfile = (length == -1 or length > MAX_INMEMORY_SIZE)

    def _get_enctype(self):
        return self.enctype

    def _get_password_cb(self):
        # Alter behaviour or the get-password-callback. For certain automatic
        # operations we don't want to ask for anything. If a password is
        # needed: too bad. And for automated tests, we want to return the
        # username + '2' as the default password.
        # (Always returning false for "caching", because it would dump
        # notices on stderr, which we don't need when testing.)
        cryptconf = PStoreCrypt.get()  # FIXME: allow this to fail?
        if cryptconf.askpass == cryptconf.ASKPASS_NEVER:
            return (lambda username, prev_was_bad: ('', False))
        elif cryptconf.askpass == cryptconf.ASKPASS_USERNAME2:
            return (lambda username, prev_was_bad: (username + '2', False))
        return None  # use default

    def decrypt_with(self, private_key):
        """
        May raise one of the CryptErrors.
        """
        # We *could* defer encryption until the first read() call is done on
        # the outputted value. But I don't see any benefits.

        # Make sure that a reused Reader provides the right info.
        if self.fp:
            if self.fpreads > 0:
                self.fp.seek(0)  # if this breaks, we didn't init fp properly
            self.fpreads += 1

        enctype = self._get_enctype()

        if enctype == 'none':
            assert private_key is None
            return self._decrypt_with_none()
        elif enctype == 'gpg':
            assert private_key is None
            return self._decrypt_with_gpg()
        elif enctype == 'sshrsa':
            return self._decrypt_with_sshrsa(private_key)
        raise NotImplementedError('unknown encryption type')

    def _decrypt_with_none(self):
        if self.fp:
            # Safe, until someone starts re-reading none-encrypted data.
            return self.fp
        else:
            # For no decryption, passing around a BytesIO is no extra effort if
            # the data already was a string.
            return BytesIO(self.data)

    def _decrypt_with_gpg(self):
        from pstorelib import gpg
        gpgcrypt = gpg.GPGCrypt()

        # Alter get-password-callback in some cases.
        password_cb = self._get_password_cb()
        if password_cb:
            def password_cb_wrapper(id_name_comment_email, key_ids,
                                    prev_was_bad):
                username = id_name_comment_email.rsplit('<', 1)[1]
                username = username.split('@', 1)[0]
                return password_cb(username, prev_was_bad)
            gpgcrypt.set_password_cb(password_cb_wrapper)

        # GPG doesn't need a decryption recipient. It finds one in your
        # secret keyring automatically. (No need to pass a private key
        # manually.)
        if self.fp:
            input = self.fp
        else:
            input = BytesIO(self.data)
        if self.use_tempfile:
            output = SpooledTemporaryFile(MAX_INMEMORY_SIZE)
        else:
            output = BytesIO()

        gpgcrypt.decrypt(input=input, output=output)
        return output

    def _decrypt_with_sshrsa(self, private_key):
        from pstorelib import sshrsa

        # Get private key from PStoreCrypt singleton if we didn't supply
        # anything.
        if not private_key:
            private_key = PStoreCrypt.get().sshrsa_privkey
        # If something was supplied, make sure it already "parsed".
        elif isinstance(private_key, basestring):
            parser = sshrsa.SSHKeyParser()
            private_key = parser.parse_private(private_key)

        kwargs = dict((i, private_key[i]) for i in 'nedpq')
        crypt = sshrsa.RSACrypt(**kwargs)

        # NOTE: We're not doing tempfiles here.. the sshrsa encryption method
        # is currently limited to very few bytes anyway.
        if self.fp:
            data = self.fp.read()
        else:
            data = self.data

        decrypted = crypt.decrypt(data)
        return BytesIO(decrypted)


def decrypts(encrypted, type, private_key=None):
    """
    Decrypt ``encrypted`` (a ``str`` (byte string) instance containing
    encrypted data) to a bytestring.

    Encryption type ``type`` must be one of 'gpg' or 'sshrsa'.

    The ``private_key`` argument is optional. It's only needed for encryption
    types that have no built-in mechanism for selecting/serving the right
    private key; i.e. the sshrsa encryption type. The expected format is some
    kind of ascii armor.

    Called ``decrypts`` and not ``decrypt`` because of its similarity with
    ``pickle.loads()`` which also takes a byte string.

    Can raise a ``CryptError``.

    NOTE: This function is not used by pstore at the moment. But it exists to
    complement encrypts which is used.
    """
    obj = CryptoReader(data=encrypted, enctype=type)
    file = obj.decrypt_with(private_key)
    return file.read()


def encrypts(unencrypted, public_key):
    """
    Encrypt ``unencrypted`` (a ``str`` (byte string) instance) as a byte string
    using ``public_key`` as public key.

    The ``public_key`` must be a human readable public key of the GPG/PGP or
    public ssh key (sshrsa) type.

    Called ``encrypts`` and not ``encrypt`` because of its similarity with
    ``pickle.dumps()`` which also returns a byte string.
    """
    obj = CryptoWriter(data=unencrypted)
    file = obj.encrypt_with(public_key)
    return file.read()


if __name__ == '__main__':
    from base64 import b64decode
    from unittest import TestCase, main

    class Test(TestCase):
        ALEX_PUBKEY = ('''
            -----BEGIN PGP PUBLIC KEY BLOCK-----
            Version: GnuPG v1.4.11 (GNU/Linux)

            mI0EULkrQAEEAKU++49M+QfiSTFJjWQ8Yyr+OKa0V90aNGbYNaGvfzlPVHNS+AwR
            PTT56yNEJAbGOFEj4aEYxa/puUJP6CZGtDvQn8FlqZ5zgKHXCRSToIVQLJjZLT3M
            Aqqkt7dCGT/nZCXASurT6zFMgR1eo3u4F3Ur5pYeLmnTUtTqC2yIrnTrABEBAAG0
            J0FsZXggQm9vbnN0cmEgKFRFU1QpIDxhbGV4QGV4YW1wbGUuY29tPoi4BBMBAgAi
            BQJQuStAAhsvBgsJCAcDAgYVCAIJCgsEFgIDAQIeAQIXgAAKCRDHl9Gl3Z236/Nw
            A/9mUbXuxDWyZsvbkqEpDJv+PjmMZ6Hk99LzSXqPomJZw0TRWmwGFRuhwl4reOXd
            fWMLfXYcrig/vNpcWmPH+1aU2yf+5wnpiGP0AzHyiarmX3smnvmMPrA8wk7F/ml3
            izheBc4+kzIWdK35QHzVzTTMjFsACQKbXAKBvO2VZb7WDLiNBFC5K0ABBAComc00
            /AE+yoO6loMzEuro1sb/HBxkH5wH8CXje3lw2TOBWf4jEW5EzPVZOpRRqfr9/2hf
            ld4GFt02ecHSe+Qk1mcaWpIUoudm/K93NH2yKaCtPFlJ+jV6qwqEG2d9LGEIVrPP
            XP9/EU75EGWOq7XOy2eIHpZOc43JOKjsuRnf9wARAQABiQE9BBgBAgAJBQJQuStA
            AhsuAKgJEMeX0aXdnbfrnSAEGQECAAYFAlC5K0AACgkQpCCsUhak5FudVAQAoqYX
            5O6Or12+jyJgCdf7oNmiuxRpMSuI2AHeSXvxFIkYrKnRyN90g9Cpg/VQtChfe9ks
            KrYzHFxS2I83Yc8j7Tw0GFqghUqFpwdzMleX6vRwBAfYzRjn+Q5lMRt5F4JUVBGN
            2fR+r9apIwpwrmztB/pqAtqW3ULXfFwOZid8ddjPnAP+LDRBMzsWTdvLcYjlh45E
            hpz5BcK9ocJTth8LfISo7eaAq80hqBcROhxegfW7o30K40TfUl5XLNBR1rV+IqSQ
            4/INWgwGTLFwwvHNNbkcNh6GbwQSMPUrV9v2MCw9K/gYHPIZl58bySm8wziElY7I
            YfZlrgW5q6LkcEoYOni58yQ=
            =XD13
            -----END PGP PUBLIC KEY BLOCK-----
        ''').strip()

        ALEX_PRIVKEY = None  # must be in gpg key chain

        JOE_PUBKEY = ('ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAAAgQDOMBVBteGqm8VFRiP'
                      'NClGy8RLYf7etIg++qA6Xo7pThorzHU0LvXK0SDeUbTSgPUAfql8pD4'
                      'hATF89HzOUs5n6YHPvAu2MMXWGfaF5DMb+5f5JbfQLhIohfllmlwWig'
                      'HinteV6uhP7xNF/4MhEJgvGwa1AP5wESR+QHtefaZDsZQ== joe')

        JOE_PRIVKEY = ('''
            -----BEGIN RSA PRIVATE KEY-----
            MIICXAIBAAKBgQDOMBVBteGqm8VFRiPNClGy8RLYf7etIg++qA6Xo7pThorzHU0L
            vXK0SDeUbTSgPUAfql8pD4hATF89HzOUs5n6YHPvAu2MMXWGfaF5DMb+5f5JbfQL
            hIohfllmlwWigHinteV6uhP7xNF/4MhEJgvGwa1AP5wESR+QHtefaZDsZQIDAQAB
            AoGAVrN1XMpEeF9EFpsl1sRuEla42/zomY7nG/7DFBSu9wxuRUie6z7gitqLNIiv
            Rmo7GpWjqjpzysXSjnseT3suVY676FZXIk9dRzJ65OBUFEvuJqpg6tvlm4Gq3mMH
            RPhWRm7xDvzQs8aH0QwDeQYr5uamcXb+4tp7/2oD73eRS4ECQQD0otMhhAdLZ6Rt
            PkbskEGE+EgTiScoNI8eyXMAg4vdB0OhFa9HP+gMXhstF4h0+rK39IvCoIbieuZJ
            74tCvWaRAkEA18QJtTx6RE8u2m7oE8HMOTkcTnAerC61xeRWOYrDzGElT2Vf+GCM
            HpoUKd3VVYA1NwXv1bGfBCXmXKaXshGalQJBANg8qJUih/QeC5hIiRaHuHJZqBQk
            kokWVD+pX07f+BwKVLwpV8KJ+YodELZ9669C/gBeV79Ud4XvjmliJN18XxECQEla
            FBLRJJ/ka4FHAw70a4FosP7ZjxPqLVHBdq7JRhdNT2nWhPHjoL8mkoRJLiWLGIxE
            MGztnanDWLVWzWt0IKECQAnz52r6SQlbAVixBzf05eTBIYxk5RUbIv0v8eC8lusD
            F5Y0uvnipvQtzpgsOxiC8c5R0j40NKfQKCVq0aPM+Jw=
            -----END RSA PRIVATE KEY-----
        ''').strip()

        def setUp(self):
            PStoreCrypt.create(askpass=PStoreCrypt.ASKPASS_USERNAME2,
                               sshrsa_privkey_file=BytesIO(self.JOE_PRIVKEY))

        def tearDown(self):
            PStoreCrypt.destroy()

        def test_decrypts_gpg(self):
            input = ('hIwDpCCsUhak5FsBA/46Fi5ubNMPWRHAiPZdoWrazqhJfg9y9+vAK2'
                     'vheNWVkpjIXvCExwzgaTmSyghbkh1SI44gybNo0RoTD/a/HxIkxN+A'
                     'EcD1+XVYOhGjUcWeS8hOk1aSmlt0caaunLUphJuN2i0v5EnomxK3ZD'
                     'qM9EyXIIsVeQxWDeDYMKEaC0Yp+9JMAWC9489HCPjwwPuyhn3Xn1Lj'
                     'X2EWOL0pSIuYrBrDba3i56z7teTTMsc69qqta5k/52OAKHzR+Evh4E'
                     'lUlI/JSIptkz8GkXalBxFLuA==')
            input = b64decode(input)
            decrypted = decrypts(input, type='gpg')
            self.assertEquals(decrypted, 'mySuperSecret!')

        def test_encrypts_gpg(self):
            input = 'mySuperSecret!'
            encrypted = encrypts(input, self.ALEX_PUBKEY)
            self.assertNotEquals(encrypted, input)  # output changes at will

        def test_encrypts_decrypts_gpg(self):
            input = "is it john doe's birth day?"
            encrypted = encrypts(input, self.ALEX_PUBKEY)
            self.assertNotEquals(encrypted, input)  # output changes at will
            decrypted = decrypts(encrypted, type='gpg')
            self.assertEquals(decrypted, input)

        def test_decrypts_sshrsa(self):
            input = ('Y9bQbUMFQ0OGvBW/iQ0PDdMy5PmsJvsxONzLKjc6uyQ3yvkZHlH/V0'
                     'OQOAc/nnMpsJfnzeLv6oZfGzRffKrU90D712f2cvpnI4/FfSGI59HS'
                     '109Q9QilXAjMLqe06P6ceOI2HtxFw6/rIDCrY9/J75+MyHJBSAZ3uw'
                     'f/lKTx1xM=')
            input = b64decode(input)
            decrypted = decrypts(input, type='sshrsa')
            self.assertEquals(decrypted, 'mySuperSecret!')

        def test_encrypts_sshrsa(self):
            input = 'mySuperSecret!'
            encrypted = encrypts(input, self.JOE_PUBKEY)
            self.assertNotEquals(encrypted, input)  # output changes at will

        def test_encrypts_decrypts_sshrsa(self):
            input = "is it john doe's birth day?"
            encrypted = encrypts(input, self.JOE_PUBKEY)
            self.assertNotEquals(encrypted, input)  # output changes at will
            decrypted = decrypts(encrypted, type='sshrsa')
            self.assertEquals(decrypted, input)

        def test_cryptowriter_none(self):
            source = 'no encryption'
            # The writer should support writing multiple times.
            writer = CryptoWriter(data=source)
            self.assertEquals(writer.encrypt_with(None).read(), source)
            self.assertEquals(writer.encrypt_with(None).read(), source)
            writer = CryptoWriter(fp=BytesIO(source))
            self.assertEquals(writer.encrypt_with(None).read(), source)
            self.assertEquals(writer.encrypt_with(None).read(), source)

        def test_cryptoreader_none(self):
            source = 'no encryption'
            reader = CryptoReader(data=source, enctype='none')
            self.assertEquals(reader.decrypt_with(None).read(), source)
            reader = CryptoReader(fp=BytesIO(source), enctype='none')
            self.assertEquals(reader.decrypt_with(None).read(), source)

        def test_cryptoreader_gpg(self):
            source = 'gpg encrypted'
            encrypted = ('hIwDpCCsUhak5FsBA/9DeiA9hg+bQUGnkE4OeW1btR0SWr2Pt'
                         'sP5O8YgEh+ZcZRB/kQlHdaV580+bttXtmU3KE30d6KWMADWqI'
                         'hlLXR3Rk7ZswsgN59u8U2IaEuAR15M8rPS7dEFz4JT3Jic+zJ'
                         'QjJstPkiWxmkdndkP2cemWEXrXuIKXs/iY45DF+/iH9JLATnn'
                         'hGZOpDpya+3jFVMkEghaWwKnvrTQLElMwuyLEf49c5uY9khQF'
                         '3k49/hGTmRbzGqrEQm0zY220OISBqVkdiFUYNDI/EKeFVaw')
            encrypted = b64decode(encrypted)
            reader = CryptoReader(data=encrypted, enctype='gpg')
            self.assertEquals(reader.decrypt_with(None).read(), source)
            reader = CryptoReader(fp=BytesIO(encrypted), enctype='gpg')
            self.assertEquals(reader.decrypt_with(None).read(), source)

        def test_cryptoreader_sshrsa(self):
            source = 'sshrsa encrypted'
            encrypted = ('lTc467y3eLwlC2wy7TOgCOzzoo9OGq3DwbVCfCsHGRf'
                         'TmIElctu8a5uQvZ5+yAsTq2AvHYBZ3s4q3aM2tpNC0s'
                         'ophQ+XPziPkZBdV04Cof9rvYXKwSuuyqbKblOjOHsRc'
                         'KguppEfSWwUnZuMRxxgoJHfvR8GeHMFE0mngGF/uyc=')
            encrypted = b64decode(encrypted)
            reader = CryptoReader(data=encrypted, enctype='sshrsa')
            self.assertEquals(reader.decrypt_with(None).read(), source)
            reader = CryptoReader(fp=BytesIO(encrypted),
                                  enctype='sshrsa')
            self.assertEquals(reader.decrypt_with(None).read(), source)

        def test_streamed_cryptowriter_and_reader(self):
            class ReadOnce(object):
                """Simulate a stream object that does not have seek."""
                def __init__(self, content):
                    self.content = content

                def read(self, size=None):
                    if size is None:
                        size = len(self.content)
                    else:
                        assert size > 0
                    ret = self.content[0:size]
                    self.content = self.content[size:]
                    return ret

            source = 'a bit of data'

            writer = CryptoWriter(fp=ReadOnce(source))
            fp1 = writer.encrypt_with(self.ALEX_PUBKEY)
            fp2 = writer.encrypt_with(self.JOE_PUBKEY)

            reader = CryptoReader(fp=fp1, enctype='gpg')
            decfp = reader.decrypt_with(None)
            self.assertEquals(decfp.read(), source)

            reader = CryptoReader(fp=fp2, enctype='sshrsa')
            decfp = reader.decrypt_with(self.JOE_PRIVKEY)
            self.assertEquals(decfp.read(), source)

    main()  # unittest.main
